"""This is the main entry point of the application."""
import os
from .api_client import get_client
from .start_analyses import start_analyses
from .download_completed_analyses import download_completed_analyses
from .config import Config
from .parse_args import parse_args

def main():
    args = parse_args()

    if args.configure:
        config = Config.singleton()
        # if you are running with --configure on the first run (you don't need to)
        # you will be prompted to configure the app by the config singleton init.
        # In that case, don't prompt the user again.
        if not config.already_prompted:
            config.run_config_prompts()

    client = get_client(api_host=os.getenv("NEFINO_API_HOST", default="https://api.nefino.li"))

    if not args.resume:
        start_analyses(client)
    else:
        download_completed_analyses(client)

if __name__ == "__main__":
    main()


from typing import Dict, List, Set 
from .schema import CoordinateInput, GeoAnalysisInput, GeoAnalysisLayerInput, GeoAnalysisObjectInput, GeoAnalysisOutputFormatInput, GeoAnalysisRequestInput, GeoAnalysisScopeInput, ScopeType
from .api_client import GeneralAvailabilityResult, LocalAvailabilityResult, build_states_list
from .journal import Journal
from .config import Config
from .access_rule_filter import AccessRuleFilter

# Place analyses require a dummy coordinate. It will be ignored in calculations.
DUMMY_COORDINATE = CoordinateInput(lon=9.0, lat=52.0)
# The API requires input of combining operations, even if they are not used.
DUMMY_OPERATIONS = []

def compose_complete_requests(general_availability: GeneralAvailabilityResult,
                              local_availability: LocalAvailabilityResult
                              ) -> Dict[str, GeoAnalysisInput]:
    """Use fetched data to build the complete requests for all available layers."""
    available_states = build_states_list(general_availability)
    requests_as_tuples = [(state, compose_single_request(state, general_availability, local_availability)) 
                          for state in available_states]
    return {state: request for (state, request) in requests_as_tuples 
            if request is not None}

def compose_layer_inputs(layers: list, local_layers: Set[str], state: str) -> List[GeoAnalysisLayerInput]:
    """Build a list of layer inputs from output lists."""
    journal = Journal.singleton()
    return [GeoAnalysisLayerInput(layer_name=layer['name'], 
                                  buffer_m=[layer['pre_buffer']]) 
            for layer in layers 
            if ((not layer.is_regional) or (layer.name in local_layers)) 
            and journal.is_newer_than_saved(layer.name, state, layer.last_update)]

def compose_single_request(state: str,
                           general_availability: GeneralAvailabilityResult,
                           local_availability: LocalAvailabilityResult
                           ) -> GeoAnalysisInput:
    """Build a single request for a given state."""
    config = Config.singleton()
    rules = AccessRuleFilter(general_availability.access_rules)
    # specify the data we want to add to the analysis
    state_local_layers = {layer.name for layer in 
                          local_availability[f'regionalLayers_{state}']}

    for skip_layer in config.skip_layers:
        state_local_layers.discard(skip_layer)

    requests_as_tuples = [(cluster, compose_layer_inputs(cluster.layers, state_local_layers, state)) 
                          for cluster in general_availability.clusters 
                          if cluster.has_access and rules.check(state, cluster.name)]

    requests =[GeoAnalysisRequestInput(cluster_name=cluster.name, layers=layers)
               for (cluster, layers) in requests_as_tuples if len(layers) > 0]

    if len(requests) == 0:
        return None
    # Specify the output format
    # TODO: this should be configurable
    output = GeoAnalysisOutputFormatInput(template_name='default', 
                                          type=config.output_format,
                                          crs=config.crs)
    # specify where the analysis should be done
    scope = GeoAnalysisScopeInput(place=state, type=ScopeType('FEDERAL_STATE'))
    # put everything together into a specification for an analysis
    spec = GeoAnalysisObjectInput(coordinate=DUMMY_COORDINATE, 
                                  output=output, 
                                  scope=scope, 
                                  requests=requests, 
                                  operations=DUMMY_OPERATIONS)
    return GeoAnalysisInput(name=f'sync_{state}', specs=spec)


from typing import Any
from .api_client import general_availability_operation, local_availability_operation, start_analyses_operation
from .compose_requests import compose_complete_requests
from .journal import Journal
from .graphql_errors import check_errors
from .parse_args import parse_args
from sgqlc.endpoint.http import HTTPEndpoint
from .download_completed_analyses import download_completed_analyses

AnalysesMutationResult = Any

def start_analyses(client: HTTPEndpoint) -> AnalysesMutationResult:
    """Starts the analyses for all updated data."""
    journal = Journal.singleton()
    args = parse_args()
    # Get information about our permissions and the general availability of layers
    general_op = general_availability_operation()
    print("Checking for layers to update...")
    general_data = client(general_op)
    check_errors(general_data)
    general_availability = (general_op + general_data)

    # Get information about the availability of layers in specific areas
    local_op = local_availability_operation(general_availability)
    local_data = client(local_op)
    check_errors(local_data)
    local_availability = (local_op + local_data)

    # Start the analyses
    analysis_inputs = compose_complete_requests(general_availability, local_availability)
    if len(analysis_inputs) == 0:
        # We can only check for layer having been unpacked already.
        # So if we're here, we've already unpacked all latest layers.
        print("✅ No layers to update. Done.")
        exit(0)
    for federal_state_key in analysis_inputs:
        print(f"Starting analysis for {federal_state_key}")
        analyses_op = start_analyses_operation({federal_state_key: analysis_inputs[federal_state_key]})
        analyses_data = client(analyses_op)
        check_errors(analyses_data)
        analyses = (analyses_op + analyses_data)

        # Add the analyses to the journal
        journal.record_analyses_requested(analyses)
        print("Analysis started.")
        download_completed_analyses(client)
    
    return(analyses)


import json
import os
import re
from datetime import datetime
from typing import Dict, Set
from .storage import get_app_directory

class Journal:
    """Handles metadata about analyses for efficient downloading."""

    # This is a singleton class. There should only be one instance of Journal.
    _instance = None

    @classmethod
    def singleton(cls):
        """Returns the singleton instance of Journal."""
        if not cls._instance:
            cls._instance = Journal()
        return cls._instance

    def __init__(self):
        if Journal._instance:
            raise Exception("Journal is a singleton class. Use Journal.singleton() to get the instance.")
        # Mapping from analysis pk to the state where the analysis was started
        self.analysis_states: Dict[str, str] = dict()

        # Mapping from layer name to where it was last updated and when
        self.layer_last_updates: Dict[str, Dict[str, datetime]] = dict()

        # Record which analyses have been successfully started
        self.synced_analyses: Set[str] = set()

        self.load_analysis_states()
        self.load_layer_last_updates()
        self.load_synced_analyses()

    def save_analysis_states(self):
        """Saves the analysis states to a file."""
        with open(os.path.join(get_app_directory(), "analysis_states.json"), "w") as f:
            json.dump(self.analysis_states, f)
    
    def load_analysis_states(self):
        """Loads the analysis states from a file."""
        try:
            with open(os.path.join(get_app_directory(), "analysis_states.json"), "r") as f:
                self.analysis_states = json.load(f)
        except FileNotFoundError:
            # we already have an empty dictionary as the field value
            print("No saved analysis states found.")
    
    def save_layer_last_updates(self):
        """Saves the layer last updates to a file."""
        with open(os.path.join(get_app_directory(), "layer_last_updates.json"), "w") as f:
            json.dump(self.layer_last_updates, f, default=lambda x: x.isoformat())

    def load_layer_last_updates(self):
        """Loads the layer last updates from a file."""
        try:
            with open(os.path.join(get_app_directory(), "layer_last_updates.json"), "r") as f:
                self.layer_last_updates = json.load(f)
                for cluster in self.layer_last_updates.values():
                    for state, timestamp in cluster.items():
                        cluster[state] = datetime.fromisoformat(timestamp) if timestamp else None
        except FileNotFoundError:
            # we already have an empty dictionary as the field value
            print("No saved layer last updates found.")
    
    def save_synced_analyses(self):
        """Saves the list of processed analyses to a file."""
        with open(os.path.join(get_app_directory(), "synced_analyses.json"), "w") as f:
            json.dump(list(self.synced_analyses), f)
    
    def load_synced_analyses(self):
        """Loads the list of processed analyses from a file."""
        try:
            with open(os.path.join(get_app_directory(), "synced_analyses.json"), "r") as f:
                self.synced_analyses = set(json.load(f))
        except FileNotFoundError:
            # we already have an empty set as the field value
            print("No saved downloaded analyses found.")

    def record_analyses_requested(self, start_analyses_result):
        """Records the analyses that have been started, and where they were started."""
        pattern = r"^startAnalysis_(?P<state>DE[1-9A-G])$"
        for alias, analysis_metadata in start_analyses_result.__dict__.items(): 
            match = re.match(pattern, alias)
            if not match:
                continue
            state = match.group("state")
            # record where the analysis was started
            self.analysis_states[analysis_metadata.pk] = state
        self.save_analysis_states()
    
    def record_layers_unpacked(self, layers: Set[str], state: str, started_at: datetime):
        """Records the layers that have been unpacked, and when they were last updated."""
        print(f"Recording layers {layers} as unpacked for state {state}")
        for layer in layers:
            if layer not in self.layer_last_updates:
                self.layer_last_updates[layer] = dict()
            self.layer_last_updates[layer][state] = started_at
        self.save_layer_last_updates()
    
    def get_state_for_analysis(self, pk: str) -> str:
        """Returns the state where the analysis was started."""
        return self.analysis_states[pk]
    
    def is_newer_than_saved(self, layer: str, state: str, timestamp: datetime) -> bool:
        """Checks if the layer needs to be unpacked."""
        if layer not in self.layer_last_updates:
            return True
        if state not in self.layer_last_updates[layer]:
            return True
        if not self.layer_last_updates[layer][state]:
            return True
        return self.layer_last_updates[layer][state] < timestamp
    
    def record_analysis_synced(self, pk: str):
        """Records that the analysis has been downloaded and unpacked."""
        self.synced_analyses.add(pk)
        self.save_synced_analyses()



from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
from .parse_args import parse_args
import json
import html

def check_errors(data: dict) -> None:
    """Check for errors in a GraphQL response."""
    args = parse_args()
    if 'errors' in data:
        if args.verbose:
            pp("<b>GraphQL operation with errors:</b> " + html.escape(json.dumps(data, indent=4)))

        if is_token_invalid(data):
            pp('<b fg="red">ERROR:</b> Invalid token. Please run <b>nefino-geosync --configure</b> and double-check your API key.')
        else:
            if not args.verbose:
                try:
                    pp("<b>Received GraphQL error from server:</b> " + html.escape(json.dumps(data['errors'], indent=4)))
                except Exception as e:
                    print(e)
                    print(data["errors"])
                pp("""<b fg="red">ERROR:</b> A GraphQL error occurred. Run with <b>--verbose</b> to see more information.
Exiting due to the above error.""")
            if args.verbose:
                pp('<b fg="red">ERROR:</b> A GraphQL error occurred. Exiting due to the above error.')

        exit(1)

def pp(to_print: str):
    print_formatted_text(HTML(to_print))

def is_token_invalid(data: dict) -> bool:
    """Check if the token is invalid."""
    try:
        if data['errors'][0]['extensions']['nefino_type'] == "AuthTokenInvalid":
            return True
    except KeyError:
        return False
    return False


import sgqlc.types
import sgqlc.types.datetime


schema = sgqlc.types.Schema()



########################################################################
# Scalars and Enumerations
########################################################################
Boolean = sgqlc.types.Boolean

class CRSType(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('EPSG_25832', 'EPSG_25833', 'EPSG_3035', 'EPSG_4326')


DateTime = sgqlc.types.datetime.DateTime

Float = sgqlc.types.Float

ID = sgqlc.types.ID

Int = sgqlc.types.Int

class OutputObjectType(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('GPKG', 'QGIS_AND_GPKG', 'QGIS_PRJ', 'SHP')


class PlaceTypeGeo(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('ADMINISTRATIVE_UNIT_GEO', 'COUNTRY', 'FEDERAL_STATE_GEO', 'LOCAL_ADMINISTRATIVE_UNITS_GEO', 'PLANNING_REGIONS_GEO')


class PlaceTypeNews(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('ADMINISTRATIVE_UNIT', 'COUNTRY', 'COUNTY', 'FEDERAL_STATE', 'LOCAL_ADMINISTRATIVE_UNITS', 'PLANNING_REGIONS')


class ScopeType(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('ADMINISTRATIVE_UNIT', 'FEDERAL_STATE', 'LOCAL_ADMINISTRATIVE_UNIT', 'PLANNING_REGION', 'POLYGON', 'RADIUS', 'SQUARE')


class Status(sgqlc.types.Enum):
    __schema__ = schema
    __choices__ = ('ERROR', 'PENDING', 'RUNNING', 'SUCCESS')


String = sgqlc.types.String

class UUID(sgqlc.types.Scalar):
    __schema__ = schema



########################################################################
# Input Objects
########################################################################
class CoordinateInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('lon', 'lat')
    lon = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='lon')
    lat = sgqlc.types.Field(sgqlc.types.non_null(Float), graphql_name='lat')


class GeoAnalysisInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('name', 'specs')
    name = sgqlc.types.Field(String, graphql_name='name')
    specs = sgqlc.types.Field(sgqlc.types.non_null('GeoAnalysisObjectInput'), graphql_name='specs')


class GeoAnalysisLayerInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('layer_name', 'buffer_m')
    layer_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='layerName')
    buffer_m = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(Int))), graphql_name='bufferM')


class GeoAnalysisObjectInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('coordinate', 'scope', 'requests', 'operations', 'output')
    coordinate = sgqlc.types.Field(sgqlc.types.non_null(CoordinateInput), graphql_name='coordinate')
    scope = sgqlc.types.Field(sgqlc.types.non_null('GeoAnalysisScopeInput'), graphql_name='scope')
    requests = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('GeoAnalysisRequestInput'))), graphql_name='requests')
    operations = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('GeoAnalysisOperationInput'))), graphql_name='operations')
    output = sgqlc.types.Field(sgqlc.types.non_null('GeoAnalysisOutputFormatInput'), graphql_name='output')


class GeoAnalysisOperationInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('operation_name', 'input')
    operation_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='operationName')
    input = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(String)), graphql_name='input')


class GeoAnalysisOutputFormatInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('template_name', 'type', 'crs')
    template_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='templateName')
    type = sgqlc.types.Field(sgqlc.types.non_null(OutputObjectType), graphql_name='type')
    crs = sgqlc.types.Field(sgqlc.types.non_null(CRSType), graphql_name='crs')


class GeoAnalysisRequestInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('cluster_name', 'layers')
    cluster_name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='clusterName')
    layers = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null(GeoAnalysisLayerInput))), graphql_name='layers')


class GeoAnalysisScopeInput(sgqlc.types.Input):
    __schema__ = schema
    __field_names__ = ('type', 'radius', 'sides', 'polygon', 'place')
    type = sgqlc.types.Field(sgqlc.types.non_null(ScopeType), graphql_name='type')
    radius = sgqlc.types.Field(Float, graphql_name='radius')
    sides = sgqlc.types.Field(Float, graphql_name='sides')
    polygon = sgqlc.types.Field(String, graphql_name='polygon')
    place = sgqlc.types.Field(String, graphql_name='place')



########################################################################
# Output Objects and Interfaces
########################################################################
class MinimalAnalysis(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('started_at', 'pk', 'status', 'url')
    started_at = sgqlc.types.Field(DateTime, graphql_name='startedAt')
    pk = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='pk')
    status = sgqlc.types.Field(sgqlc.types.non_null(Status), graphql_name='status')
    url = sgqlc.types.Field(String, graphql_name='url')


class MinimalAreasEnabled(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('all_areas_enabled', 'enabled_states')
    all_areas_enabled = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='allAreasEnabled')
    enabled_states = sgqlc.types.Field(sgqlc.types.list_of('PlaceIdentifier'), graphql_name='enabledStates')


class MinimalCluster(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('name', 'layers', 'has_access')
    name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='name')
    layers = sgqlc.types.Field(sgqlc.types.non_null(sgqlc.types.list_of(sgqlc.types.non_null('MinimalLayer'))), graphql_name='layers')
    has_access = sgqlc.types.Field(Boolean, graphql_name='hasAccess')


class MinimalGeoAccessRule(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('all_clusters_enabled', 'clusters', 'places')
    all_clusters_enabled = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='allClustersEnabled')
    clusters = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='clusters')
    places = sgqlc.types.Field(sgqlc.types.list_of(String), graphql_name='places')


class MinimalLayer(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('name', 'pre_buffer', 'last_update', 'is_regional')
    name = sgqlc.types.Field(sgqlc.types.non_null(String), graphql_name='name')
    pre_buffer = sgqlc.types.Field(sgqlc.types.non_null(Int), graphql_name='preBuffer')
    last_update = sgqlc.types.Field(DateTime, graphql_name='lastUpdate')
    is_regional = sgqlc.types.Field(sgqlc.types.non_null(Boolean), graphql_name='isRegional')


class Mutation(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('start_analysis',)
    start_analysis = sgqlc.types.Field(MinimalAnalysis, graphql_name='startAnalysis', args=sgqlc.types.ArgDict((
        ('inputs', sgqlc.types.Arg(sgqlc.types.non_null(GeoAnalysisInput), graphql_name='inputs', default=None)),
))
    )


class PlaceIdentifier(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('place_id', 'place_type')
    place_id = sgqlc.types.Field(sgqlc.types.non_null(ID), graphql_name='placeId')
    place_type = sgqlc.types.Field(sgqlc.types.non_null(PlaceTypeNews), graphql_name='placeType')


class Query(sgqlc.types.Type):
    __schema__ = schema
    __field_names__ = ('analysis_metadata', 'allowed_analysis_areas', 'clusters', 'regional_layers', 'access_rules')
    analysis_metadata = sgqlc.types.Field(sgqlc.types.list_of(MinimalAnalysis), graphql_name='analysisMetadata', args=sgqlc.types.ArgDict((
        ('analysis_id', sgqlc.types.Arg(UUID, graphql_name='analysisId', default=None)),
))
    )
    allowed_analysis_areas = sgqlc.types.Field(MinimalAreasEnabled, graphql_name='allowedAnalysisAreas')
    clusters = sgqlc.types.Field(sgqlc.types.list_of(MinimalCluster), graphql_name='clusters')
    regional_layers = sgqlc.types.Field(sgqlc.types.list_of(MinimalLayer), graphql_name='regionalLayers', args=sgqlc.types.ArgDict((
        ('place_type', sgqlc.types.Arg(sgqlc.types.non_null(PlaceTypeGeo), graphql_name='placeType', default=None)),
        ('place_id', sgqlc.types.Arg(sgqlc.types.non_null(String), graphql_name='placeId', default=None)),
))
    )
    access_rules = sgqlc.types.Field(sgqlc.types.list_of(MinimalGeoAccessRule), graphql_name='accessRules')



########################################################################
# Unions
########################################################################

########################################################################
# Schema Entry Points
########################################################################
schema.query_type = Query
schema.mutation_type = Mutation
schema.subscription_type = None

